var RefHelper, argsNormalizer, arrayMutators, hasKeys, merge, mutators, pathParser, specHelper, transaction, _ref;
var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; }, __slice = Array.prototype.slice;
transaction = require('./transaction');
pathParser = require('./pathParser');
specHelper = require('./specHelper');
_ref = require('./util'), merge = _ref.merge, hasKeys = _ref.hasKeys;
argsNormalizer = require('./argsNormalizer');
mutators = require('./mutators');
arrayMutators = mutators.array;
module.exports = RefHelper = function(model) {
  this._model = model;
  this._adapter = model._adapter;
  this._setup();
};
RefHelper.prototype = {
  isPointer: function(val) {
    return !!(val && val.$r);
  },
  isKeyPath: function(path, obj) {
    var found;
    if (!obj) {
      throw new Error('Missing obj');
    }
    found = this._adapter.lookup("$keys." + path + ".$", false, {
      obj: obj
    }).obj;
    return found !== void 0;
  },
  isPathPointedTo: function(path, obj) {
    var found;
    if (!obj) {
      throw 'Missing obj';
    }
    found = this._adapter.lookup("$refs." + path + ".$", false, {
      obj: obj
    }).obj;
    return found !== void 0;
  },
  _setup: function() {
    var adapter, indexesInArgs, method, normalizeArgs, refHelper, _ref2, _results;
    refHelper = this;
    adapter = this._adapter;
    adapter.__set = adapter.set;
    adapter.set = function(path, value, ver, options) {
      var eachNode, maybeIndexRefs, out;
      if (options == null) {
        options = {};
      }
      out = null;
      options.obj || (options.obj = this._data);
      maybeIndexRefs = __bind(function(_path, _value) {
        if (refHelper.isPointer(_value)) {
          refHelper.$indexRefs(_path, _value.$r, _value.$k, _value.$t, ver, options);
        }
        if (_path === path) {
          out = this.__set(path, value, ver, options);
        }
        return refHelper.updateRefsForKey(_path, ver, options);
      }, this);
      if (Object !== (value != null ? value.constructor : void 0)) {
        maybeIndexRefs(path, value);
      } else {
        eachNode = function(path, value, callback) {
          var nodePath, prop, val, _results;
          callback(path, value);
          _results = [];
          for (prop in value) {
            val = value[prop];
            nodePath = "" + path + "." + prop;
            _results.push(Object === (val != null ? val.constructor : void 0) ? eachNode(nodePath, val, callback) : callback(nodePath, val));
          }
          return _results;
        };
        eachNode(path, value, function(nodePath, nodeValue) {
          return maybeIndexRefs(nodePath, nodeValue);
        });
      }
      return out;
    };
    adapter.__del = adapter.del;
    adapter.del = function(path, ver, options) {
      var out;
      if (options == null) {
        options = {};
      }
      options.obj || (options.obj = this._data);
      out = this.__del(path, ver, options);
      if (refHelper.isPathPointedTo(path, options.obj)) {
        refHelper.cleanupPointersTo(path, options);
      }
      return out;
    };
    _results = [];
    for (method in arrayMutators) {
      _ref2 = arrayMutators[method], normalizeArgs = _ref2.normalizeArgs, indexesInArgs = _ref2.indexesInArgs;
      adapter['__' + method] = adapter[method];
      _results.push(adapter[method] = (function(method, normalizeArgs, indexesInArgs) {
        return function() {
          var methodArgs, newIndexes, options, out, path, ver, _ref3;
          _ref3 = normalizeArgs.apply(null, arguments), path = _ref3.path, methodArgs = _ref3.methodArgs, ver = _ref3.ver, options = _ref3.options;
          options.obj || (options.obj = this._data);
          if (indexesInArgs) {
            newIndexes = indexesInArgs(methodArgs).map(function(index) {
              return refHelper.arrRefIndex(index, path, options.obj);
            });
            indexesInArgs(methodArgs, newIndexes);
          }
          out = this['__' + method].apply(this, [path].concat(__slice.call(methodArgs), [ver], [options]));
          refHelper.updateRefsForKey(path, ver, options);
          return out;
        };
      })(method, normalizeArgs, indexesInArgs));
    }
    return _results;
  },
  arrRefIndex: function(start, path, obj) {
    var arr, i, mem, startIndex, _len;
    if ('number' === typeof start) {
      return start;
    }
    arr = this._adapter.lookup(path, true, {
      obj: obj
    }).obj;
    if (this.isArrayRef(path, obj)) {
      startIndex = arr.length;
      for (i = 0, _len = arr.length; i < _len; i++) {
        mem = arr[i];
        if (mem.id === start.id || parseInt(mem.id, 10) === parseInt(start.id, 10)) {
          return startIndex = i;
        }
      }
    }
    startIndex = arr.indexOf(start.id);
    if (startIndex !== -1) {
      return startIndex;
    }
    startIndex = arr.indexOf(parseInt(start.id, 10));
    if (startIndex !== -1) {
      return startIndex;
    }
    return arr.indexOf(start.id.toString());
  },
  ref: function(ref, key) {
    if (key != null) {
      return {
        $r: ref,
        $k: key
      };
    } else {
      return {
        $r: ref
      };
    }
  },
  arrayRef: function(ref, key) {
    return {
      $r: ref,
      $k: key,
      $t: 'array'
    };
  },
  $indexRefs: function(path, ref, key, type, ver, options) {
    var adapter, entry, keyOptions, keyVal, oldKey, oldRefObj, oldRefOptions, refHelper, refs, refsKey, refsKeys;
    adapter = this._adapter;
    refHelper = this;
    oldRefOptions = merge({
      dontFollowLastRef: true
    }, options);
    oldRefObj = adapter.lookup(path, false, oldRefOptions).obj;
    if (key) {
      entry = [ref, key];
      if (type) {
        entry.push(type);
      }
      adapter.lookup("$keys." + key + ".$", true, options).obj[path] = entry;
      keyVal = adapter.lookup(key, false, options).obj;
      if (type === void 0 && keyVal === void 0) {
        return;
      }
      if (type === 'array') {
        keyOptions = merge({
          array: true
        }, options);
        keyVal = adapter.lookup(key, true, keyOptions).obj;
        refsKeys = keyVal.map(function(keyValMem) {
          return ref + '.' + keyValMem;
        });
        this._removeOld$refs(oldRefObj, path, ver, options);
        return refsKeys.forEach(function(refsKey) {
          return refHelper._update$refs(refsKey, path, ref, key, type, options);
        });
      }
      refsKey = ref + '.' + keyVal;
    } else {
      if (oldRefObj && (oldKey = oldRefObj.$k)) {
        refs = adapter.lookup("$keys." + oldKey + ".$", false, options).obj;
        if (refs && refs[path]) {
          delete refs[path];
          if (!hasKeys(refs, {
            ignore: specHelper.reserved
          })) {
            adapter.del("$keys." + oldKey, ver, options);
          }
        }
      }
      refsKey = ref;
    }
    this._removeOld$refs(oldRefObj, path, ver, options);
    return this._update$refs(refsKey, path, ref, key, type, options);
  },
  _removeOld$refs: function(oldRefObj, path, ver, options) {
    var oldKey, oldKeyVal, oldRef, refHelper;
    if (oldRefObj && (oldRef = oldRefObj.$r)) {
      if (oldKey = oldRefObj.$k) {
        oldKeyVal = this._adapter.lookup(oldKey, false, options).obj;
      }
      if (oldKey && (oldRefObj.$t === 'array')) {
        refHelper = this;
        oldKeyVal.forEach(function(oldKeyMem) {
          return refHelper._removeFrom$refs(oldRef, oldKeyMem, path, ver, options);
        });
        return this._removeFrom$refs(oldRef, void 0, path, ver, options);
      } else {
        return this._removeFrom$refs(oldRef, oldKeyVal, path, ver, options);
      }
    }
  },
  _removeFrom$refs: function(ref, key, path, ver, options) {
    var refEntries, refWithKey;
    if (key) {
      refWithKey = ref + '.' + key;
    }
    refEntries = this._adapter.lookup("$refs." + refWithKey + ".$", false, options).obj;
    if (!refEntries) {
      return;
    }
    delete refEntries[path];
    if (!hasKeys(refEntries, {
      ignore: specHelper.reserved
    })) {
      return this._adapter.del("$refs." + ref, ver, options);
    }
  },
  _update$refs: function(refsKey, path, ref, key, type, options) {
    var entry, out;
    entry = [ref, key];
    if (type) {
      entry.push(type);
    }
    out = this._adapter.lookup("$refs." + refsKey + ".$", true, options);
    return out.obj[path] = entry;
  },
  updateRefsForKey: function(path, ver, options) {
    var refs;
    if (refs = this._adapter.lookup("$keys." + path + ".$", false, options).obj) {
      this._eachValidRef(refs, options.obj, __bind(function(path, ref, key, type) {
        return this.$indexRefs(path, ref, key, type, ver, options);
      }, this));
    }
    return this.eachValidRefPointingTo(path, options.obj, __bind(function(pointingPath, targetPathRemainder, ref, key, type) {
      return this.updateRefsForKey(pointingPath + '.' + targetPathRemainder, ver, options);
    }, this));
  },
  _eachValidRef: function(refs, obj, callback) {
    var fastLookup, key, o, path, ref, type, _ref2, _results;
    if (obj == null) {
      obj = this._adapter._data;
    }
    fastLookup = pathParser.fastLookup;
    _results = [];
    for (path in refs) {
      _ref2 = refs[path], ref = _ref2[0], key = _ref2[1], type = _ref2[2];
      if (path === '_proto') {
        continue;
      }
      o = fastLookup(path, obj);
      o = this._adapter.lookup(path, false, {
        obj: obj,
        dontFollowLastRef: true
      }).obj;
      _results.push(o && o.$r === ref && o.$k == key ? callback(path, ref, key, type) : delete refs[path]);
    }
    return _results;
  },
  _eachRefSetPointingTo: function(path, refs, fn) {
    var i, prop, props, refPos, refSet, _results;
    i = 0;
    refPos = refs;
    props = path.split('.');
    _results = [];
    while (prop = props[i++]) {
      if (!(refPos = refPos[prop])) {
        return;
      }
      _results.push((refSet = refPos.$) ? fn(refSet, props.slice(i).join('.'), prop) : void 0);
    }
    return _results;
  },
  eachValidRefPointingTo: function(targetPath, obj, fn) {
    var adapter, refs, self;
    self = this;
    adapter = self._adapter;
    if (!(refs = adapter.lookup('$refs', false, {
      obj: obj
    }).obj)) {
      return;
    }
    return self._eachRefSetPointingTo(targetPath, refs, function(refSet, targetPathRemainder, possibleIndex) {
      return self._eachValidRef(refSet, obj, function(pointingPath, ref, key, type) {
        if (type === 'array') {
          targetPathRemainder = possibleIndex + '.' + targetPathRemainder;
        }
        return fn(pointingPath, targetPathRemainder, ref, key, type);
      });
    });
  },
  eachArrayRefKeyedBy: function(path, obj, fn) {
    var key, ref, refSet, refs, type, _ref2, _results;
    if (!(refs = this._adapter.lookup("$keys", false, {
      obj: obj
    }).obj)) {
      return;
    }
    refSet = (path + '.$').split('.').reduce(function(refSet, prop) {
      return refSet && refSet[prop];
    }, refs);
    if (!refSet) {
      return;
    }
    _results = [];
    for (path in refSet) {
      _ref2 = refSet[path], ref = _ref2[0], key = _ref2[1], type = _ref2[2];
      _results.push(type === 'array' ? fn(path, ref, key) : void 0);
    }
    return _results;
  },
  notifyPointersTo: function(targetPath, obj, method, args, ignoreRoots) {
    if (ignoreRoots == null) {
      ignoreRoots = [];
    }
    this.eachValidRefPointingTo(targetPath, obj, __bind(function(pointingPath, targetPathRemainder, ref, key, type) {
      var id, index, rest, _ref2;
      if (type !== 'array') {
        if (this._alreadySeen(pointingPath, ref, ignoreRoots)) {
          return;
        }
        if (targetPathRemainder) {
          pointingPath += '.' + targetPathRemainder;
        }
      } else if (targetPathRemainder) {
        _ref2 = targetPathRemainder.split('.'), id = _ref2[0], rest = 2 <= _ref2.length ? __slice.call(_ref2, 1) : [];
        index = this._toIndex(key, id, obj);
        if (index !== -1) {
          pointingPath += '.' + index;
          if (rest.length) {
            pointingPath += '.' + rest.join('.');
          }
        }
      }
      return this._model.emit(method, [pointingPath].concat(__slice.call(args)));
    }, this));
    return this.eachArrayRefKeyedBy(targetPath, obj, __bind(function(pointingPath, ref, key) {
      var arrayMemberArgs, firstArgs, ns, _ref2;
      _ref2 = (mutators.basic[method] || mutators.array[method]).splitArgs(args), firstArgs = _ref2[0], arrayMemberArgs = _ref2[1];
      if (arrayMemberArgs) {
        ns = this._adapter.lookup(ref, false, {
          obj: obj
        }).obj;
        arrayMemberArgs = arrayMemberArgs.map(function(arg) {
          return ns[arg] || ns[parseInt(arg, 10)];
        });
      }
      args = firstArgs.concat(arrayMemberArgs);
      return this._model.emit(method, [pointingPath].concat(__slice.call(args)));
    }, this));
  },
  _toIndex: function(arrayRefKey, id, obj) {
    var index, keyArr;
    keyArr = this._adapter.lookup(arrayRefKey, false, {
      array: true,
      obj: obj
    }).obj;
    index = keyArr.indexOf(id);
    if (index === -1) {
      return keyArr.indexOf(parseInt(id, 10));
    }
    return index;
  },
  _alreadySeen: function(pointingPath, ref, ignoreRoots) {
    var alreadySeen;
    alreadySeen = ignoreRoots.some(function(root) {
      return root === pointingPath.substr(0, root.length);
    });
    if (alreadySeen) {
      return true;
    }
    ignoreRoots.push(ref);
    return false;
  },
  cleanupPointersTo: function(path, options) {
    var adapter, key, keyMem, keyVal, model, pointingPath, ref, refs, _ref2, _results;
    adapter = this._adapter;
    refs = adapter.lookup("$refs." + path + ".$", false, options).obj;
    if (refs === void 0) {
      return;
    }
    model = this._model;
    _results = [];
    for (pointingPath in refs) {
      _ref2 = refs[pointingPath], ref = _ref2[0], key = _ref2[1];
      keyVal = key && adapter.lookup(key, false, options).obj;
      _results.push(keyVal && specHelper.isArray(keyVal) ? (keyMem = path.substr(ref.length + 1, pointingPath.length), adapter.remove(key, keyVal.indexOf(keyMem), 1, null, options)) : void 0);
    }
    return _results;
  },
  dereferenceTxn: function(txn, obj) {
    var $k, $r, args, method, path, remainingProps, _ref2, _ref3, _ref4;
    method = transaction.method(txn);
    args = transaction.args(txn);
    path = transaction.path(txn);
    if (method in arrayMutators) {
      if (_ref2 = this.isArrayRef(path, obj), $r = _ref2.$r, $k = _ref2.$k, _ref2) {
        this._model._cache.invalidateSpecModelCache();
        args[0] = path = $k;
        if (arrayMutators[method].argsToForeignKeys) {
          args = arrayMutators[method].argsToForeignKeys(args, path, $r);
        }
      } else {
        _ref3 = this._adapter.lookup(path, false, {
          obj: obj
        }), obj = _ref3.obj, path = _ref3.path, remainingProps = _ref3.remainingProps;
        if (obj === void 0 && (remainingProps != null ? remainingProps.length : void 0)) {
          args[0] = [path].concat(remainingProps).join('.');
        } else {
          args[0] = path;
        }
      }
      return txn;
    }
    _ref4 = this._adapter.lookup(path, false, {
      obj: obj
    }), obj = _ref4.obj, path = _ref4.path, remainingProps = _ref4.remainingProps;
    if (obj === void 0 && (remainingProps != null ? remainingProps.length : void 0)) {
      args[0] = [path].concat(remainingProps).join('.');
    } else {
      args[0] = path;
    }
    return txn;
  },
  isArrayRef: function(path, data) {
    var $k, $r, $t, options, refObj;
    options = {
      obj: data,
      dontFollowLastRef: true
    };
    refObj = this._adapter.lookup(path, false, options).obj;
    if (refObj === void 0) {
      return false;
    }
    $r = refObj.$r, $k = refObj.$k, $t = refObj.$t;
    if ($t !== 'array') {
      return false;
    }
    $k && ($k = this.dereferencedPath($k, data));
    return {
      $r: $r,
      $k: $k
    };
  },
  isRefObj: function(obj) {
    return '$r' in obj;
  },
  dereferencedPath: function(path, data) {
    var meta;
    meta = this._adapter.lookup(path, false, {
      obj: data,
      returnMeta: true
    });
    return meta.path;
  },
  denormalizePath: function(path, data) {
    var obj, remainingProps, _ref2;
    _ref2 = this._adapter.lookup(path, false, {
      obj: data
    }), path = _ref2.path, obj = _ref2.obj, remainingProps = _ref2.remainingProps;
    if (obj || !(remainingProps != null ? remainingProps.length : void 0)) {
      return path;
    }
    return [path].concat(remainingProps).join('.');
  }
};